import abc
from six import with_metaclass
from .mixins import EvaluatePermissionsMixin


class BasePermission(with_metaclass(abc.ABCMeta)):
    """
    A base class from which all permission classes should inherit.
    """

    @abc.abstractmethod
    def has_permission(self, request, view):
        pass

    @abc.abstractmethod
    def has_object_permission(self, request, view, obj):
        pass


class BaseComponent(with_metaclass(abc.ABCMeta, BasePermission)):
    """
    A base class from which all component classes should inherit.
    """

    def __invert__(self):
        return Not(self)

    def __or__(self, component):
        return Or(self, component)

    def __and__(self, component):
        return And(self, component)


class Not(EvaluatePermissionsMixin, BaseComponent):

    def __init__(self, component):
        self.component = component

    def evaluate_permissions(self, permission_name, *args, **kwargs):
        return not getattr(self.component, permission_name)(*args, **kwargs)


class Or(EvaluatePermissionsMixin, BaseComponent):

    def __init__(self, *components):
        self.components = components

    def evaluate_permissions(self, permission_name, *args, **kwargs):
        return any(getattr(component, permission_name)(*args, **kwargs)
                   for component in self.components)


class And(EvaluatePermissionsMixin, BaseComponent):

    def __init__(self, *components):
        self.components = components

    def evaluate_permissions(self, permission_name, *args, **kwargs):
        return all(getattr(component, permission_name)(*args, **kwargs)
                   for component in self.components)



__version__ = '1.0.0'
__title__ = 'django-rest-action-permissions'
__description__ = 'Action based permissions for Django REST Framework.'
__url__ = 'https://github.com/apirobot/django-rest-action-permissions'
__author__ = 'Denis Orehovsky'
__author_email__ = 'denis.orehovsky@gmail.com'
__license__ = 'MIT License'



class EvaluatePermissionsMixin:

    def has_permission(self, request, view):
        return self.evaluate_permissions('has_permission', request, view)

    def has_object_permission(self, request, view, obj):
        return self.evaluate_permissions(
            'has_object_permission', request, view, obj
        )



import inspect
import functools
import operator
from django.core.exceptions import ImproperlyConfigured
from rest_framework import permissions
from .base import BaseComponent, BasePermission
from .mixins import EvaluatePermissionsMixin


class ActionPermission(EvaluatePermissionsMixin, BasePermission):

    def evaluate_permissions(
        self, permission_name, request, view, *args, **kwargs
    ):
        perms = self._get_required_permissions(request, view)
        return getattr(perms, permission_name)(request, view, *args, **kwargs)

    def _get_required_permissions(self, request, view):
        action = self._get_action(view.action)
        action_perms_attr = '{}_perms'.format(action)
        perms = self._get_permissions_by_attr(action_perms_attr)
        if perms is None:
            if request.method in permissions.SAFE_METHODS:
                read_perms_attr = 'read_perms'
                perms = self._get_permissions_by_attr(read_perms_attr)
                if perms is None:
                    self._raise_attr_error(
                        general_attr=read_perms_attr,
                        action_attr=action_perms_attr
                    )
            else:
                write_perms_attr = 'write_perms'
                perms = self._get_permissions_by_attr(write_perms_attr)
                if perms is None:
                    self._raise_attr_error(
                        general_attr=write_perms_attr,
                        action_attr=action_perms_attr
                    )

        global_perms_attr = 'global_perms'
        global_perms = self._get_permissions_by_attr(global_perms_attr)
        if global_perms is not None:
            perms = global_perms & perms

        enough_perms_attr = 'enough_perms'
        enough_perms = self._get_permissions_by_attr(enough_perms_attr)
        if enough_perms is not None:
            perms = enough_perms | perms

        return perms

    def _get_permissions_by_attr(self, attr):
        if not hasattr(self, attr):
            return None
        perms = getattr(self, attr)
        perms = self._validate_permissions(
            perms=perms, attr=attr
        )
        return perms

    def _get_action(self, action):
        meta = getattr(self, 'Meta', None)
        partial_update_is_update = getattr(
            meta, 'partial_update_is_update', True
        )
        if action == 'partial_update' and partial_update_is_update:
            return 'update'
        return action

    def _validate_permissions(self, perms, attr):
        if isinstance(perms, BaseComponent):
            return perms
        elif isinstance(perms, (tuple, list)):
            validated_perms = (self._validate_permissions(perm, attr)
                               for perm in perms)
            # If we have a list of components, then we need to convert them
            # into one single component using `And` operator.
            return functools.reduce(operator.and_, validated_perms)
        elif inspect.isclass(perms) and issubclass(perms, BaseComponent):
            return perms()
        else:
            self._raise_validation_error(attr=attr)

    def _raise_attr_error(self, action_attr, general_attr):
        raise RuntimeError(
            '`{cls}` class must declare `{cls}.{action_attr}` or '
            '`{cls}.{general_attr}` attributes'.format(
                cls=self.__class__.__name__,
                action_attr=action_attr,
                general_attr=general_attr
            )
        )

    def _raise_validation_error(self, attr):
        raise ImproperlyConfigured(
            '`{cls}` class has invalid permission definition in '
            '`{cls}.{attr}` attribute'.format(
                cls=self.__class__.__name__,
                attr=attr
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import sys
from shutil import rmtree
from setuptools import setup, Command

here = os.path.abspath(os.path.dirname(__file__))


# Long description
with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    readme = '\n' + f.read()


about = {}
# Get meta-data from __version__.py
with open(os.path.join(here, 'rest_action_permissions', '__version__.py')) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(
            sys.executable
        ))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


# Where the magic happens:
setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    license=about['__license__'],
    packages=[
        'rest_action_permissions',
    ],
    install_requires=[
        'django',
        'djangorestframework',
        'six',
    ],
    include_package_data=True,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    cmdclass={
        'upload': UploadCommand,
    },
)



from .base import BaseComponent


class ActionPermissionComponent(BaseComponent):

    def has_permission(self, request, view):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True

    def has_object_permission(self, request, view, obj):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True


class AllowAny(ActionPermissionComponent):
    """
    Allow any access.
    """

    def has_permission(self, request, view):
        return True

    def has_object_permission(self, request, view, obj):
        return True


class DenyAll(ActionPermissionComponent):
    """
    Deny any access.
    """

    def has_permission(self, request, view):
        return False

    def has_object_permission(self, request, view, obj):
        return False


class IsAuthenticated(ActionPermissionComponent):
    """
    Allows access only to authenticated users.
    """

    def has_permission(self, request, view):
        return request.user and request.user.is_authenticated

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view)


class IsAdminUser(ActionPermissionComponent):
    """
    Allows access only to admin users.
    """

    def has_permission(self, request, view):
        return (request.user
                and request.user.is_authenticated
                and request.user.is_staff)

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view)


class IsSuperUser(ActionPermissionComponent):
    """
    Allows access only to superusers.
    """

    def has_permission(self, request, view):
        return (request.user
                and request.user.is_authenticated
                and request.user.is_superuser)

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view)
